# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.logger import LOGGER
from rp2.rp2_decimal import FIAT_DECIMAL_MASK, ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class InTransaction(AbstractTransaction):
    @classmethod
    def type_check(cls, name: str, instance: AbstractEntry) -> "InTransaction":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(
        self,
        configuration: Configuration,
        timestamp: str,
        asset: str,
        exchange: str,
        holder: str,
        transaction_type: str,
        spot_price: RP2Decimal,
        crypto_in: RP2Decimal,
        fiat_fee: RP2Decimal,
        fiat_in_no_fee: Optional[RP2Decimal] = None,
        fiat_in_with_fee: Optional[RP2Decimal] = None,
        unique_id: Optional[int] = None,
        notes: Optional[str] = None,
    ) -> None:
        super().__init__(configuration, timestamp, asset, transaction_type, spot_price, unique_id, notes)

        self.__exchange: str = configuration.type_check_exchange("exchange", exchange)
        self.__holder: str = configuration.type_check_holder("holder", holder)
        self.__crypto_in: RP2Decimal = configuration.type_check_positive_decimal("crypto_in", crypto_in, non_zero=True)
        self.__fiat_fee: RP2Decimal = configuration.type_check_positive_decimal("fiat_fee", fiat_fee)

        # Fiat in with/without fee are optional. They can be derived from crypto in, spot price and fiat fee, however some exchanges
        # provide them anyway. If they are provided use them as given by the exchange, if not compute them.
        self.__fiat_in_no_fee: RP2Decimal
        self.__fiat_in_with_fee: RP2Decimal
        if fiat_in_no_fee is None:
            self.__fiat_in_no_fee = self.__crypto_in * self.spot_price
        else:
            self.__fiat_in_no_fee = configuration.type_check_positive_decimal("fiat_in_no_fee", fiat_in_no_fee, non_zero=True)
        if fiat_in_with_fee is None:
            self.__fiat_in_with_fee = self.__fiat_in_no_fee + self.__fiat_fee
        else:
            self.__fiat_in_with_fee = configuration.type_check_positive_decimal("fiat_in_with_fee", fiat_in_with_fee, non_zero=True)

        if spot_price == ZERO:
            raise RP2ValueError(f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.unique_id}): parameter 'spot_price' cannot be 0")
        if (
            self.transaction_type != TransactionType.BUY
            and self.transaction_type != TransactionType.GIFT
            and self.transaction_type != TransactionType.DONATE
            and not self.transaction_type.is_earn_type()
        ):
            raise RP2ValueError(f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.unique_id}): invalid transaction type {self.transaction_type}")

        # If the values provided by the exchange doesn't match the computed one, log a warning.
        if not RP2Decimal.is_equal_within_precision(self.__crypto_in * self.spot_price, self.__fiat_in_no_fee, FIAT_DECIMAL_MASK):
            LOGGER.warning(
                "%s %s (%s, id %s): crypto_in * spot_price != fiat_in_no_fee: %f != %f",
                self.asset,
                type(self).__name__,
                self.timestamp,
                self.unique_id,
                self.__crypto_in * self.spot_price,
                self.__fiat_in_no_fee,
            )
        if not RP2Decimal.is_equal_within_precision(self.__fiat_in_with_fee, self.__fiat_in_no_fee + self.__fiat_fee, FIAT_DECIMAL_MASK):
            LOGGER.warning(
                "%s %s (%s, id %s): fiat_in_with_fee != fiat_in_no_fee + fiat_fee: %f != %f",
                self.asset,
                type(self).__name__,
                self.timestamp,
                self.unique_id,
                self.__fiat_in_with_fee,
                self.__fiat_in_no_fee + self.__fiat_fee,
            )

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        self.configuration.type_check_positive_int("indent", indent)
        self.configuration.type_check_bool("repr_format", repr_format)
        if extra_data and not isinstance(extra_data, List):
            raise RP2TypeError(f"Parameter 'extra_data' is not of type List: {extra_data}")

        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str
        class_specific_data = [
            f"exchange={stringify(self.exchange)}",
            f"holder={stringify(self.holder)}",
            f"transaction_type={stringify(self.transaction_type)}",
            f"spot_price={self.spot_price:.4f}",
            f"crypto_in={self.crypto_in:.8f}",
            f"fiat_fee={self.fiat_fee:.4f}",
            f"fiat_in_no_fee={self.fiat_in_no_fee:.4f}",
            f"fiat_in_with_fee={self.fiat_in_with_fee:.4f}",
            f"is_taxable={stringify(self.is_taxable())}",
            f"fiat_taxable_amount={self.fiat_taxable_amount:.4f}",
        ]
        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def exchange(self) -> str:
        return self.__exchange

    @property
    def holder(self) -> str:
        return self.__holder

    @property
    def crypto_in(self) -> RP2Decimal:
        return self.__crypto_in

    @property
    def fiat_in_no_fee(self) -> RP2Decimal:
        return self.__fiat_in_no_fee

    @property
    def fiat_in_with_fee(self) -> RP2Decimal:
        return self.__fiat_in_with_fee

    @property
    def fiat_fee(self) -> RP2Decimal:
        return self.__fiat_fee

    @property
    def crypto_taxable_amount(self) -> RP2Decimal:
        if self.is_taxable():
            return self.__crypto_in
        return ZERO

    @property
    def fiat_taxable_amount(self) -> RP2Decimal:
        if self.is_taxable():
            # InTransactions that have is_taxable() set to True should not have any fees, but we return fiat_in_with_fee conservatively
            return self.fiat_in_with_fee
        return ZERO

    @property
    def crypto_deduction(self) -> RP2Decimal:
        return ZERO

    @property
    def fiat_deduction(self) -> RP2Decimal:
        return ZERO

    @property
    def crypto_balance_change(self) -> RP2Decimal:
        return self.crypto_in

    @property
    def fiat_balance_change(self) -> RP2Decimal:
        return self.fiat_in_with_fee

    def is_taxable(self) -> bool:
        return self.transaction_type.is_earn_type()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from datetime import date, datetime
from typing import Dict, List, Optional, Set

from rp2.abstract_entry import AbstractEntry
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.entry_types import EntrySetType
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.out_transaction import OutTransaction
from rp2.rp2_error import RP2TypeError, RP2ValueError


class AbstractEntrySet:
    def __init__(
        self,
        configuration: Configuration,
        entry_set_type: str,
        asset: str,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ) -> None:
        self.__configuration: Configuration = Configuration.type_check("configuration", configuration)
        self.__entry_set_type: EntrySetType = EntrySetType.type_check_from_string("entry_set_type", entry_set_type)
        self.__asset: str = configuration.type_check_asset("asset", asset)
        if not isinstance(from_date, date):
            raise RP2TypeError("Parameter 'from_date' is not of type date")
        self._from_date: date = from_date
        if not isinstance(to_date, date):
            raise RP2TypeError("Parameter 'to_date' is not of type date")
        self._to_date: date = to_date

        self._entry_list: List[AbstractEntry] = []  # List for sorting
        self._entry_set: Set[AbstractEntry] = set()  # Set for fast search (at the cost of extra memory)
        self._entry_to_parent: Dict[AbstractEntry, Optional[AbstractEntry]] = {}
        self.__is_sorted: bool = False

    # @classmethod
    # def join(cls, entry_sets: List["AbstractEntrySet"]) -> "AbstractEntrySet":
    #     # pylint: disable=protected-access
    #     if not isinstance(entry_sets, List):
    #         raise RP2TypeError("Parameter 'entry_sets' is not a List")
    #     if len(entry_sets) < 1:
    #         raise RP2ValueError("List 'entry_sets' has no elements")
    #     result: Optional[AbstractEntrySet] = None
    #     for entry_set in entry_sets:
    #         result = AbstractEntrySet(entry_set.configuration, entry_set.entry_set_type, entry_set.asset, entry_set.from_date, entry_set.to_date) if not result else result
    #         if entry_set.configuration != result.configuration or entry_set.entry_set_type != result.entry_set_type or entry_set.asset != result.asset or entry_set.from_date != result.from_date or entry_set.to_date != result.to_date:
    #             raise RP2ValueError("Joining entry sets failed: parameters are incompatible.")
    #         result._entry_list.extend(entry_set._entry_list)
    #         result._entry_set.update(entry_set._entry_set)
    #     result._force_sort()
    #     return result

    def duplicate(self, from_date: date = MIN_DATE, to_date: date = MAX_DATE) -> "AbstractEntrySet":
        # pylint: disable=protected-access
        result: AbstractEntrySet = copy(self)
        result._from_date = from_date
        result._to_date = to_date
        # Force sort to recompute fields that are affected by time filter
        result._force_sort()
        return result

    def __str__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}:")
        output.append(f"  configuration={str(self.__configuration.configuration_path)}")
        output.append(f"  entry_set_type={str(self.entry_set_type)}")
        output.append(f"  asset={str(self.asset)}")
        output.append(f"  from_date={self._from_date if self._from_date > MIN_DATE else 'non-specified'}")
        output.append(f"  to_date={self._to_date if self._to_date < MAX_DATE else 'non-specified'}")
        output.append("  entries=")
        for entry in self:
            parent: Optional[AbstractEntry]
            parent = self.get_parent(entry)
            output.append(entry.to_string(indent=2, repr_format=False, extra_data=[f"parent={str(parent.unique_id if parent else None)}"]))
        return "\n".join(output)

    def __repr__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}(")
        output.append(f"configuration={repr(self.__configuration.configuration_path)}")
        output.append(f", entry_set_type={repr(self.entry_set_type)}")
        output.append(f", asset={repr(self.asset)}")
        output.append(f", from_date={repr(self._from_date) if self._from_date > MIN_DATE else 'non-specified'}")
        output.append(f", to_date={repr(self._to_date) if self._to_date < MAX_DATE else 'non-specified'}")
        output.append(", entries=[")
        for entry in self:
            parent: Optional[AbstractEntry]
            parent = self.get_parent(entry)
            output.append(entry.to_string(indent=2, repr_format=True, extra_data=[f"parent={repr(parent.unique_id if parent else None)}"]))
        output.append("]")
        output.append(")")
        return "".join(output)

    @property
    def configuration(self) -> Configuration:
        return self.__configuration

    @property
    def entry_set_type(self) -> EntrySetType:
        return self.__entry_set_type

    @property
    def asset(self) -> str:
        return self.__asset

    @property
    def from_date(self) -> date:
        return self._from_date

    @property
    def to_date(self) -> date:
        return self._to_date

    @property
    def count(self) -> int:
        return len(self._entry_list)

    def add_entry(self, entry: AbstractEntry) -> None:
        AbstractEntry.type_check("entry", entry)

        if entry.asset != self.asset:
            raise RP2ValueError(f"Attempting to add a {entry.asset} entry to a {self.asset} set")
        if self.entry_set_type == EntrySetType.IN and not isinstance(entry, InTransaction):
            raise RP2TypeError(f"Attempting to add a {entry.__class__.__name__} to a set of type IN")
        if self.entry_set_type == EntrySetType.INTRA and not isinstance(entry, IntraTransaction):
            raise RP2TypeError(f"Attempting to add a {entry.__class__.__name__} to a set of type INTRA")
        if self.entry_set_type == EntrySetType.OUT and not isinstance(entry, OutTransaction):
            raise RP2TypeError(f"Attempting to add a {entry.__class__.__name__} to a set of type OUT")
        if entry in self._entry_set:
            raise RP2ValueError(f"Entry already added: {entry}")

        self._entry_list.append(entry)
        self._entry_set.add(entry)
        self.__is_sorted = False

    def is_empty(self) -> bool:
        return self.count == 0

    def _validate_entry(self, entry: AbstractEntry) -> None:
        AbstractEntry.type_check("entry", entry)
        if entry not in self._entry_set:
            raise RP2ValueError(f"Unknown entry:\n{entry}")

    def get_parent(self, entry: AbstractEntry) -> Optional[AbstractEntry]:
        self._validate_entry(entry)
        self._check_sort()
        return self._entry_to_parent[entry]

    def _sort_entries(self) -> None:
        # Sort entries by date, then add parent
        self._entry_list.sort(key=_entry_sort_key)
        parent: Optional[AbstractEntry] = None
        for entry in self._entry_list:
            self._entry_to_parent[entry] = parent
            parent = entry

    def _check_sort(self) -> None:
        if not self.__is_sorted:
            self._sort_entries()
            self.__is_sorted = True

    def _force_sort(self) -> None:
        self.__is_sorted = False
        self._check_sort()

    def __iter__(self) -> "EntrySetIterator":
        self._check_sort()
        return EntrySetIterator(self)


class EntrySetIterator:
    def __init__(self, entry_set: AbstractEntrySet) -> None:
        self.__entry_set: AbstractEntrySet = entry_set
        self.__entry_set_size: int = self.__entry_set.count
        self.__index: int = 0

    def __next__(self) -> AbstractEntry:
        result: Optional[AbstractEntry] = None
        while self.__index < self.__entry_set_size:
            result = self.__entry_set._entry_list[self.__index]  # pylint: disable=protected-access
            self.__index += 1
            if result.timestamp.date() > self.__entry_set.to_date:
                raise StopIteration(self)
            if result.timestamp.date() >= self.__entry_set.from_date:
                return result
        raise StopIteration(self)


def _entry_sort_key(entry: AbstractEntry) -> datetime:
    return entry.timestamp



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from decimal import Decimal
from typing import Generic, Optional, TypeVar

KeyType = TypeVar("KeyType", int, datetime, Decimal, float, str)
ValueType = TypeVar("ValueType")


class AVLNode(Generic[KeyType, ValueType]):
    def __init__(self, key: KeyType, value: ValueType):
        self.__key: KeyType = key
        self.__value: ValueType = value
        self.__height: int = 1

        self.__left: Optional[AVLNode[KeyType, ValueType]] = None
        self.__right: Optional[AVLNode[KeyType, ValueType]] = None

    def __repr__(self) -> str:
        return f"{self.to_string()}"

    def to_string(self) -> str:
        return (
            f"{type(self).__name__}(key={repr(self.key)}, "
            f"value={repr(self.value)}, height={repr(self.height)}, "
            f"left={self.left.to_string() if self.left else None}, "
            f"right={self.right.to_string() if self.right else None})"
        )

    @property
    def key(self) -> KeyType:
        return self.__key

    @property
    def value(self) -> ValueType:
        return self.__value

    @property
    def left(self) -> "Optional[AVLNode[KeyType, ValueType]]":
        return self.__left

    @left.setter
    def left(self, node: "Optional[AVLNode[KeyType, ValueType]]") -> None:
        self.__left = node

    @property
    def right(self) -> "Optional[AVLNode[KeyType, ValueType]]":
        return self.__right

    @right.setter
    def right(self, node: "Optional[AVLNode[KeyType, ValueType]]") -> None:
        self.__right = node

    @property
    def height(self) -> int:
        return self.__height

    @height.setter
    def height(self, height: int) -> None:
        self.__height = height


class AVLTree(Generic[KeyType, ValueType]):
    def __init__(self) -> None:
        self.__root: Optional[AVLNode[KeyType, ValueType]] = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}(root={repr(self.__root)})"

    def find_max_value_less_than(self, key: KeyType) -> Optional[ValueType]:
        result: Optional[AVLNode[KeyType, ValueType]]
        result = self.find_max_node_less_than_at_node(self.__root, key) if self.__root else None
        return result.value if result is not None else None

    def insert_node(self, key: KeyType, value: ValueType) -> None:
        new_node: AVLNode[KeyType, ValueType]
        new_node = self.insert_node_at_node(self.__root, key, value)
        if self.__root != new_node:
            self.__root = new_node

    @staticmethod
    def find_max_node_less_than_at_node(root: AVLNode[KeyType, ValueType], key: KeyType) -> Optional[AVLNode[KeyType, ValueType]]:
        current_node: Optional[AVLNode[KeyType, ValueType]] = root
        result: Optional[AVLNode[KeyType, ValueType]] = None
        while current_node is not None:
            current_key: KeyType = current_node.key
            try:
                if current_key > key:
                    current_node = current_node.left
                elif current_key < key:
                    result = current_node
                    current_node = current_node.right
                elif current_key == key:
                    result = current_node
                    break
            except TypeError as exc:
                raise TypeError("AVLTree: keys are not comparable") from exc
        return result

    def insert_node_at_node(self, root: Optional[AVLNode[KeyType, ValueType]], key: KeyType, value: ValueType) -> AVLNode[KeyType, ValueType]:
        if not root:
            return AVLNode(key, value)

        if key < root.key:
            root.left = self.insert_node_at_node(root.left, key, value)
        else:
            root.right = self.insert_node_at_node(root.right, key, value)

        root.height = 1 + max(self._get_height(root.left), self._get_height(root.right))

        balance_factor: int = self._get_balance_factor(root)
        if balance_factor > 1:
            # Disable mypy on the next few lines: it complains about root.left possibly being None (and therefore not having
            # attribute "key"). However since balance_factor is > 1 root.left is guaranteed not to be None.
            if key < root.left.key:  # type: ignore
                return self._rotate_right(root)
            root.left = self._rotate_left(root.left)  # type: ignore
            return self._rotate_right(root)
        if balance_factor < -1:
            # Disable mypy on the next few lines: it complains about root.right possibly being None (and therefore not having
            # attribute "key"). However since balance_factor is < -1 root.right is guaranteed not to be None.
            if key > root.right.key:  # type: ignore
                return self._rotate_left(root)
            root.right = self._rotate_right(root.right)  # type: ignore
            return self._rotate_left(root)

        return root

    # Rotation implementation based on: https://en.wikipedia.org/wiki/Tree_rotation
    def _rotate_left(self, root: AVLNode[KeyType, ValueType]) -> AVLNode[KeyType, ValueType]:
        # Disable mypy on the next few lines: it complains that variables possibly being None (and therefore not having accessible
        # attributes). However unless there is a bug, this should never occur.
        pivot: Optional[AVLNode[KeyType, ValueType]] = root.right
        root.right = pivot.left  # type: ignore
        pivot.left = root  # type: ignore
        root.height = 1 + max(self._get_height(root.left), self._get_height(root.right))  # type: ignore
        pivot.height = 1 + max(self._get_height(pivot.left), self._get_height(pivot.right))  # type: ignore
        return pivot  # type: ignore

    def _rotate_right(self, root: AVLNode[KeyType, ValueType]) -> AVLNode[KeyType, ValueType]:
        # Disable mypy on the next few lines: it complains that variables possibly being None (and therefore not having accessible
        # attributes). However unless there is a bug, this should never occur.
        pivot: Optional[AVLNode[KeyType, ValueType]] = root.left
        root.left = pivot.right  # type: ignore
        pivot.right = root  # type: ignore
        root.height = 1 + max(self._get_height(root.left), self._get_height(root.right))  # type: ignore
        pivot.height = 1 + max(self._get_height(pivot.left), self._get_height(pivot.right))  # type: ignore
        return pivot  # type: ignore

    @staticmethod
    def _get_height(root: Optional[AVLNode[KeyType, ValueType]]) -> int:
        return root.height if root else 0

    def _get_balance_factor(self, root: AVLNode[KeyType, ValueType]) -> int:
        return self._get_height(root.left) - self._get_height(root.right) if root else 0

    @property
    def root(self) -> Optional[AVLNode[KeyType, ValueType]]:
        return self.__root
